/*
  wavesmooth.cpp

  image smoothing using the discrete wavlet transformation (DWT)

  Copyright (C) 2004, Patrick Rammelt <rammelt@cs.tu-berlin.de>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/* 
   compile:
   g++ -Wall -g -O3 wavesmooth.cpp -ltiff -ljpeg -o ../bin/Wave
*/

#include "photoimage.cpp"
#include "wavelet.cpp"


// TEST
int main (int argc, char **argv) 
{
  // wavelet-filters
  class dwtAnalyzer dwta (dwt_Debauchy97_low,
			  dwt_Debauchy97_high,
			  dwt_Debauchy97_size); 
  
  class dwtNoise dwtn (dwt_Debauchy97_low,
		       dwt_Debauchy97_high,
		       dwt_Debauchy97_size);

  class dwtBlur dwtb (dwt_Debauchy97_low,
		      dwt_Debauchy97_high,
		      dwt_Debauchy97_size);
	    
  // defaults
  char *in = NULL, *out = "smoothed.tif";
  int start = 0, quality = 75;
  long size  = 512, overlap = 64;
  long depth = 0, shifts = 3, skip = 0, radius = 5, blur = 1;
  double level1 = 128, delta = 64;
  bool showflt = 0, shownoise = 0, showweights = 0;

  // get options
  for (long i=1; i<argc; i++) {
    if      (!strcmp(argv[i],"-in")     && i+1<argc) in     = argv[++i];
    else if (!strcmp(argv[i],"-out")    && i+1<argc) out    = argv[++i];
    else if (!strcmp(argv[i],"-depth")  && i+1<argc) depth  = atoi(argv[++i]); 
    else if (!strcmp(argv[i],"-shifts") && i+1<argc) shifts = atoi(argv[++i]); 
    else if (!strcmp(argv[i],"-skip")   && i+1<argc) skip   = atoi(argv[++i]); 
    else if (!strcmp(argv[i],"-level")  && i+1<argc) level1 = atof(argv[++i]);
    else if (!strcmp(argv[i],"-blur")   && i+1<argc) blur   = atoi(argv[++i]);
    else if (!strcmp(argv[i],"-delta")  && i+1<argc) delta  = atof(argv[++i]);
    else if (!strcmp(argv[i],"-radius") && i+1<argc) radius = atoi(argv[++i]);
    else if (!strcmp(argv[i],"-rgb"))                start  = 0;
    else if (!strcmp(argv[i],"-grey"))               start  = 3;
    else if (!strcmp(argv[i],"-quality")&& i+1<argc) quality= atoi(argv[++i]);
    else if (!strcmp(argv[i],"-size")   && i+1<argc) size   = atoi(argv[++i]);
    else if (!strcmp(argv[i],"-overlap")&& i+1<argc) overlap= atoi(argv[++i]);
    else if (!strcmp(argv[i],"-filter"))             showflt     = 1;
    else if (!strcmp(argv[i],"-noise"))              shownoise   = 1;
    else if (!strcmp(argv[i],"-weights"))            showweights = 1;
    else if (!strcmp(argv[i],"-debug")) showflt = shownoise = showweights = 1;
    else if (!strcmp(argv[i],"-help")) {
      printf ("Wave\n  -in infile  [-out outfile]\n");
      printf ("  [-depth depth]  [-shifts shifts]  [-skip skip]\n");
      printf ("  [-level 100%%-level]  [-delta blur-100%%-level]\n");
      printf ("  [-blur blur-depth]  [-radius blur-radius]\n");
      printf ("  [{-rgb|-grey}]  [-quality jpeg_quality]\n"); 
      printf ("  [-size tile-size]  [-overlap tile-overlap]\n"); 
      printf ("  [-filter]  [-noise]  [-weights]  [-debug]\n"); 
      exit (0); }
    else {
      printf ("Unknown option: \"%s\" - try Wave -help\n",argv[i]);
      exit (0); }
  }

  // check
  if (in==NULL) { 
    printf ("No input file\n"); 
    exit (0); 
  }

  // modify/set options
  depth = dwtn.getDepth (size,depth);
  dwtn.setParams (level1,shifts,skip);
  dwtb.setParams (delta,radius);

  // show options
  printf ("input        = %s\n",in);
  printf ("output       = %s\n",out);
  printf ("size/overlap = %ld/%ld\n",size,overlap);
  printf ("level 100%%   = %g\n",level1);
  printf ("delta        = %g\n",delta);
  printf ("blur-depth   = %ld\n",blur);
  printf ("radius       = %ld\n",radius);
  printf ("shifts       = %ld\n",shifts);
  printf ("skip         = %ld\n",skip);
  printf ("depth        = %ld\n",depth);
  printf ("quality      = %d (jpeg only)\n",quality);

  // get image
  printf ("Read image\n"); fflush (stdout);
  class color rgb;
  class photoimage inimg;
  inimg.read_jpeg (in);
  long w = inimg.width();
  long h = inimg.height();
  class photoimage outimg(w,h);
  class photoimage fltimg(size,size);
  class photoimage noiseimg(size,size);
  class photoimage weightimg(size,size*2);

  // signals
  double *sig0    = new double[size*size];
  double *sig1    = new double[size*size];
  double *sig2    = new double[size*size];
  double *weights = new double[size*size*2]; 

  // init signal (image might be smaller)
  for (long i=size*size-1; i>=0; sig0[i]=sig1[i]=0, i--);

  // denoise all channels (r,g,b)
  unsigned int  stop = MAX(3,start+1); 
  for (unsigned int channel=start; channel<stop; channel++) {
    
    // copy color-channel-information to sig1
    printf("channel: %s\n",(channel==0 ? "red" : 
			    (channel==1 ? "green" : 
			     (channel==2 ? "blue" : "grey")))); fflush(stdout);
    
    for (long y0=0; y0<h-overlap; y0+=size-overlap) {
      for (long x0=0; x0<w-overlap; x0+=size-overlap) {
	printf ("  tile: (%ld,%ld) - (%ld,%ld)\n",x0,y0,x0+size-1,y0+size-1);

	for (long y=0; y<MIN(h-y0,size); y++) {
	  for (long x=0; x<MIN(w-x0,size); x++) {
	    rgb = inimg(x0+x,y0+y);
	    if (channel>2)
	      sig0[x+y*size] = sig1[x+y*size] = rgb.grey() -128.0;
	    else
	      sig0[x+y*size] = sig1[x+y*size] = rgb.channel(channel) - 128.0;
	  }
	}
  
	// analysis step
	double *siga[2] = { sig1 , sig2 };
	double *resa = dwta.multi(siga,size,2,depth); 
  	
	// fill filtered image 
	if (showflt) {
	  for (long y=0; y<size; y++) {
	    for (long x=0; x<size; x++) {
	      rgb = fltimg(x,y);
	      if (channel>2)
		rgb.grey(resa[x+y*size]+128.0);
	      else
		rgb.channel(resa[x+y*size]+128.0,channel);
	      fltimg(x,y,rgb);
	    }
	  }
	}

	// get weightings for denoising
	double *sign[3] = { resa , siga[1], weights };
	double *resn = dwtn.multi(sign,size,2,depth);

	// fill noise-image
	if (shownoise) {
	  for (long y=0; y<size; y++) {
	    for (long x=0; x<size; x++) {
	      rgb = noiseimg(x,y);
	      if (channel>2)
		rgb.grey(resn[x+y*size]+128.0);
	      else
		rgb.channel(resn[x+y*size]+128.0,channel);
	      noiseimg(x,y,rgb);
	    }
	  }
	}
	
	// fill weights-image
	if (showweights) {
	  for (long y=0; y<size*2; y++) {
	    for (long x=0; x<size; x++) {
	      rgb = weightimg(x,y);
	      if (channel>2)
		rgb.grey(weights[x+y*size]*256);
	      else
		rgb.channel(weights[x+y*size]*256,channel);
	      weightimg(x,y,rgb);
	    }
	  }
	}

	// denoise + selective blur
	dwtb.denoise (sig0, resn, size*size, 1);


	// second analysis step
	double *siga2[2] = { sig0 , sig1 };
	double *resa2 = dwta.multi(siga2,size,2,blur); 

	// blur
	double *sigb[3] = { resa2, sig2, weights };
	double *resb = dwtb.multi(sigb,size,2,blur);

	//const long len[2] = { MIN(size,w), MIN(size,h) };
	//const long step[2] = { 1, size };
	//dwtb.blur (sig0, sig1, &(weights[size*size]), 2, len, step);

	// copy channel back to image
	for (long dy=(y0 ? overlap : 0), y=0; y<MIN(h-y0,size); y++) {
	  for (long dx=(x0 ? overlap : 0), x=0; x<MIN(w-x0,size); x++) {
	    rgb = outimg(x+x0,y+y0);
	    const double c  = (channel>2 ? rgb.grey() : rgb.channel(channel));
	    const double wx = (double)MAX(dx-x,0)/(double)overlap;
	    const double wy = (double)MAX(dy-y,0)/(double)overlap;
	    const double w  = MAX(wx,wy);
	    if (channel>2) 
	      rgb.grey((1.0-w) * (resb[x+y*size] + 128) + w*c);
	    else 
	      rgb.channel((1.0-w) * (resb[x+y*size] + 128) + w*c, channel); 
	    outimg(x+x0,y+y0,rgb);
	  }
	}

      }
    }

  }

  // write result(s)
  printf ("Write image\n"); fflush (stdout);
  if (showflt)     fltimg.write_tiff ("filtered.tif");
  if (shownoise)   noiseimg.write_tiff ("noise.tif");
  if (showweights) weightimg.write_tiff ("weights.tif");

  int isjpeg = 0; char *pnt;
  if ((pnt = rindex(out,'.'))!=NULL) isjpeg = !strcmp(pnt,".jpg");
  if (isjpeg) outimg.write_jpeg (out,quality);
  else        outimg.write_tiff (out);

}
